@@ -1,9 +1,12 @@
+from __future__ import absolute_import
+from __future__ import division
+from __future__ import print_function
 import numpy as np
 import pandas as pd
 
-from .pycompat import basestring, iteritems, suppress, dask_array_type, bytes_type
+from .pycompat import basestring, suppress, dask_array_type, OrderedDict
 from . import formatting
-from .utils import SortedKeysDict, not_implemented
+from .utils import SortedKeysDict, not_implemented, Frozen
 
 
 class ImplementsArrayReduce(object):
@@ -29,6 +32,13 @@ class ImplementsArrayReduce(object):
             and 'axis' arguments can be supplied. If neither are supplied, then
             `{name}` is calculated over axes."""
 
+    _cum_extra_args_docstring = \
+        """dim : str or sequence of str, optional
+            Dimension over which to apply `{name}`.
+        axis : int or sequence of int, optional
+            Axis over which to apply `{name}`. Only one of the 'dim'
+            and 'axis' arguments can be supplied."""
+
 
 class ImplementsDatasetReduce(object):
     @classmethod
@@ -48,68 +58,20 @@ class ImplementsDatasetReduce(object):
 
     _reduce_extra_args_docstring = \
         """dim : str or sequence of str, optional
-            Dimension(s) over which to apply `func`.  By default `func` is
+            Dimension(s) over which to apply `{name}`.  By default `{name}` is
             applied over all dimensions."""
 
-
-class ImplementsRollingArrayReduce(object):
-    @classmethod
-    def _reduce_method(cls, func):
-        def wrapped_func(self, **kwargs):
-            return self.reduce(func, **kwargs)
-        return wrapped_func
-
-    @classmethod
-    def _bottleneck_reduce(cls, func):
-        def wrapped_func(self, **kwargs):
-            from .dataarray import DataArray
-
-            if isinstance(self.obj.data, dask_array_type):
-                raise NotImplementedError(
-                    'Rolling window operation does not work with dask arrays')
-
-            # bottleneck doesn't allow min_count to be 0, although it should
-            # work the same as if min_count = 1
-            if self.min_periods is not None and self.min_periods == 0:
-                min_count = self.min_periods + 1
-            else:
-                min_count = self.min_periods
-
-            values = func(self.obj.data, window=self.window,
-                          min_count=min_count, axis=self._axis_num)
-
-            result = DataArray(values, self.obj.coords)
-
-            if self.center:
-                result = self._center_result(result)
-
-            return result
-        return wrapped_func
-
-    @classmethod
-    def _bottleneck_reduce_without_min_count(cls, func):
-        def wrapped_func(self, **kwargs):
-            from .dataarray import DataArray
-
-            if self.min_periods is not None:
-                raise ValueError('Rolling.median does not accept min_periods')
-
-            if isinstance(self.obj.data, dask_array_type):
-                raise NotImplementedError(
-                    'Rolling window operation does not work with dask arrays')
-
-            values = func(self.obj.data, window=self.window, axis=self._axis_num)
-
-            result = DataArray(values, self.obj.coords)
-
-            if self.center:
-                result = self._center_result(result)
-
-            return result
-        return wrapped_func
+    _cum_extra_args_docstring = \
+        """dim : str or sequence of str, optional
+            Dimension over which to apply `{name}`.
+        axis : int or sequence of int, optional
+            Axis over which to apply `{name}`. Only one of the 'dim'
+            and 'axis' arguments can be supplied."""
 
 
 class AbstractArray(ImplementsArrayReduce, formatting.ReprMixin):
+    """Shared base class for DataArray and Variable."""
+
     def __bool__(self):
         return bool(self.values)
 
@@ -172,6 +134,18 @@ class AbstractArray(ImplementsArrayReduce, formatting.ReprMixin):
             raise ValueError("%r not found in array dimensions %r" %
                              (dim, self.dims))
 
+    @property
+    def sizes(self):
+        """Ordered mapping from dimension names to lengths.
+
+        Immutable.
+
+        See also
+        --------
+        Dataset.sizes
+        """
+        return Frozen(OrderedDict(zip(self.dims, self.shape)))
+
 
 class AttrAccessMixin(object):
     """Mixin class that allows getting keys with attribute access
@@ -181,7 +155,7 @@ class AttrAccessMixin(object):
     @property
     def _attr_sources(self):
         """List of places to look-up items for attribute-style access"""
-        return [self, self.attrs]
+        return []
 
     def __getattr__(self, name):
         if name != '__setstate__':
@@ -217,7 +191,61 @@ class AttrAccessMixin(object):
         return sorted(set(dir(type(self)) + extra_attrs))
 
 
+def get_squeeze_dims(xarray_obj, dim):
+    """Get a list of dimensions to squeeze out.
+    """
+    if dim is None:
+        dim = [d for d, s in xarray_obj.sizes.items() if s == 1]
+    else:
+        if isinstance(dim, basestring):
+            dim = [dim]
+        if any(xarray_obj.sizes[k] > 1 for k in dim):
+            raise ValueError('cannot select a dimension to squeeze out '
+                             'which has length greater than one')
+    return dim
+
+
 class BaseDataObject(AttrAccessMixin):
+    """Shared base class for Dataset and DataArray."""
+
+    def squeeze(self, dim=None, drop=False):
+        """Return a new object with squeezed data.
+
+        Parameters
+        ----------
+        dim : None or str or tuple of str, optional
+            Selects a subset of the length one dimensions. If a dimension is
+            selected with length greater than one, an error is raised. If
+            None, all length one dimensions are squeezed.
+        drop : bool, optional
+            If ``drop=True``, drop squeezed coordinates instead of making them
+            scalar.
+
+        Returns
+        -------
+        squeezed : same type as caller
+            This object, but with with all or a subset of the dimensions of
+            length 1 removed.
+
+        See Also
+        --------
+        numpy.squeeze
+        """
+        dims = get_squeeze_dims(self, dim)
+        return self.isel(drop=drop, **{d: 0 for d in dims})
+
+    def get_index(self, key):
+        """Get an index for a dimension, with fall-back to a default RangeIndex
+        """
+        if key not in self.dims:
+            raise KeyError(key)
+
+        try:
+            return self.indexes[key]
+        except KeyError:
+            # need to ensure dtype=int64 in case range is empty on Python 2
+            return pd.Index(range(self.sizes[key]), name=key, dtype=np.int64)
+
     def _calc_assign_results(self, kwargs):
         results = SortedKeysDict()
         for k, v in kwargs.items():
@@ -228,8 +256,10 @@ class BaseDataObject(AttrAccessMixin):
         return results
 
     def assign_coords(self, **kwargs):
-        """Assign new coordinates to this object, returning a new object
-        with all the original data in addition to the new coordinates.
+        """Assign new coordinates to this object.
+
+        Returns a new object with all the original data in addition to the new
+        coordinates.
 
         Parameters
         ----------
@@ -262,6 +292,29 @@ class BaseDataObject(AttrAccessMixin):
         data.coords.update(results)
         return data
 
+    def assign_attrs(self, *args, **kwargs):
+        """Assign new attrs to this object.
+
+        Returns a new object equivalent to self.attrs.update(*args, **kwargs).
+
+        Parameters
+        ----------
+        args : positional arguments passed into ``attrs.update``.
+        kwargs : keyword arguments passed into ``attrs.update``.
+
+        Returns
+        -------
+        assigned : same type as caller
+            A new object with the new attrs in addition to the existing data.
+
+        See also
+        --------
+        Dataset.assign
+        """
+        out = self.copy(deep=False)
+        out.attrs.update(*args, **kwargs)
+        return out
+
     def pipe(self, func, *args, **kwargs):
         """
         Apply func(self, *args, **kwargs)
@@ -326,7 +379,7 @@ class BaseDataObject(AttrAccessMixin):
 
         Parameters
         ----------
-        group : str, DataArray or Coordinate
+        group : str, DataArray or IndexVariable
             Array whose unique values should be used to group this array. If a
             string, must be the name of a variable contained in this dataset.
         squeeze : boolean, optional
@@ -340,9 +393,7 @@ class BaseDataObject(AttrAccessMixin):
             A `GroupBy` object patterned after `pandas.GroupBy` that can be
             iterated over in the form of `(unique_value, grouped_array)` pairs.
         """
-        if isinstance(group, basestring):
-            group = self[group]
-        return self.groupby_cls(self, group, squeeze=squeeze)
+        return self._groupby_cls(self, group, squeeze=squeeze)
 
     def groupby_bins(self, group, bins, right=True, labels=None, precision=3,
                      include_lowest=False, squeeze=True):
@@ -353,7 +404,7 @@ class BaseDataObject(AttrAccessMixin):
 
         Parameters
         ----------
-        group : str, DataArray or Coordinate
+        group : str, DataArray or IndexVariable
             Array whose binned values should be used to group this array. If a
             string, must be the name of a variable contained in this dataset.
         bins : int or array of scalars
@@ -391,12 +442,10 @@ class BaseDataObject(AttrAccessMixin):
         ----------
         .. [1] http://pandas.pydata.org/pandas-docs/stable/generated/pandas.cut.html
         """
-        if isinstance(group, basestring):
-            group = self[group]
-        return self.groupby_cls(self, group, squeeze=squeeze, bins=bins,
-                                cut_kwargs={'right': right, 'labels': labels,
-                                            'precision': precision,
-                                            'include_lowest': include_lowest})
+        return self._groupby_cls(self, group, squeeze=squeeze, bins=bins,
+                                 cut_kwargs={'right': right, 'labels': labels,
+                                             'precision': precision,
+                                             'include_lowest': include_lowest})
 
     def rolling(self, min_periods=None, center=False, **windows):
         """
@@ -425,8 +474,8 @@ class BaseDataObject(AttrAccessMixin):
         rolling : type of input argument
         """
 
-        return self.rolling_cls(self, min_periods=min_periods,
-                                center=center, **windows)
+        return self._rolling_cls(self, min_periods=min_periods,
+                                 center=center, **windows)
 
     def resample(self, freq, dim, how='mean', skipna=None, closed=None,
                  label=None, base=0, keep_attrs=False):
@@ -496,7 +545,7 @@ class BaseDataObject(AttrAccessMixin):
         group = DataArray(dim, [(RESAMPLE_DIM, dim)], name=RESAMPLE_DIM)
         time_grouper = pd.TimeGrouper(freq=freq, how=how, closed=closed,
                                       label=label, base=base)
-        gb = self.groupby_cls(self, group, grouper=time_grouper)
+        gb = self._groupby_cls(self, group, grouper=time_grouper)
         if isinstance(how, basestring):
             f = getattr(gb, how)
             if how in ['first', 'last']:
@@ -519,9 +568,11 @@ class BaseDataObject(AttrAccessMixin):
         ----------
         cond : boolean DataArray or Dataset
         other : unimplemented, optional
-            Unimplemented placeholder for compatibility with future numpy / pandas versions
+            Unimplemented placeholder for compatibility with future
+            numpy / pandas versions
         drop : boolean, optional
-            Coordinate labels that only correspond to NA values should be dropped
+            Coordinate labels that only correspond to NA values should be
+            dropped
 
         Returns
         -------
@@ -553,7 +604,8 @@ class BaseDataObject(AttrAccessMixin):
           * y        (y) int64 0 1 2 3 4
         """
         if other is not None:
-            raise NotImplementedError("The optional argument 'other' has not yet been implemented")
+            raise NotImplementedError("The optional argument 'other' has not "
+                                      "yet been implemented")
 
         if drop:
             from .dataarray import DataArray
@@ -571,31 +623,35 @@ class BaseDataObject(AttrAccessMixin):
             clip = dict(zip(clipcond.dims, [np.unique(adim)
                                             for adim in np.nonzero(clipcond.values)]))
             outcond = cond.isel(**clip)
-            outobj = self.sel(**outcond.indexes)
+            indexers = {dim: outcond.get_index(dim) for dim in outcond.dims}
+            outobj = self.sel(**indexers)
         else:
             outobj = self
             outcond = cond
 
-        return outobj._where(outcond)
+        # preserve attributes
+        out = outobj._where(outcond)
+        out._copy_attrs_from(self)
+        return out
 
-    # this has no runtime function - these are listed so IDEs know these methods
-    # are defined and don't warn on these operations
-    __lt__ = __le__ =__ge__ = __gt__ = __add__ = __sub__ = __mul__ = \
-    __truediv__ = __floordiv__ = __mod__ = __pow__ = __and__  = __xor__ = \
-    __or__ = __div__ = __eq__ = __ne__ = not_implemented
+    def close(self):
+        """Close any files linked to this object
+        """
+        if self._file_obj is not None:
+            self._file_obj.close()
+        self._file_obj = None
 
+    def __enter__(self):
+        return self
 
-def squeeze(xarray_obj, dims, dim=None):
-    """Squeeze the dims of an xarray object."""
-    if dim is None:
-        dim = [d for d, s in iteritems(dims) if s == 1]
-    else:
-        if isinstance(dim, basestring):
-            dim = [dim]
-        if any(dims[k] > 1 for k in dim):
-            raise ValueError('cannot select a dimension to squeeze out '
-                             'which has length greater than one')
-    return xarray_obj.isel(**dict((d, 0) for d in dim))
+    def __exit__(self, exc_type, exc_value, traceback):
+        self.close()
+
+    # this has no runtime function - these are listed so IDEs know these
+    # methods are defined and don't warn on these operations
+    __lt__ = __le__ = __ge__ = __gt__ = __add__ = __sub__ = __mul__ = \
+        __truediv__ = __floordiv__ = __mod__ = __pow__ = __and__ = __xor__ = \
+        __or__ = __div__ = __eq__ = __ne__ = not_implemented
 
 
 def _maybe_promote(dtype):
@@ -634,54 +690,70 @@ def _get_fill_value(dtype):
     return fill_value
 
 
-def _full_like_dataarray(arr, keep_attrs=False, fill_value=None):
-    """empty DataArray"""
-    from .dataarray import DataArray
-
-    attrs = arr.attrs if keep_attrs else {}
-
-    if fill_value is None:
-        values = np.empty_like(arr)
-    elif fill_value is True:
-        dtype, fill_value = _maybe_promote(arr.dtype)
-        values = np.full_like(arr, fill_value=fill_value, dtype=dtype)
-    else:
-        dtype, _ = _maybe_promote(np.array(fill_value).dtype)
-        values = np.full_like(arr, fill_value=fill_value, dtype=dtype)
-
-    return DataArray(values, dims=arr.dims, coords=arr.coords, attrs=attrs)
-
-
-def _full_like(xray_obj, keep_attrs=False, fill_value=None):
+def full_like(other, fill_value, dtype=None):
     """Return a new object with the same shape and type as a given object.
 
     Parameters
     ----------
-    xray_obj : DataArray or Dataset
-        Return a full object with the same shape/dims/coords/attrs.
-            `func` is calculated over all dimension for each group item.
-    keep_attrs : bool, optional
-        If True, the datasets's attributes (`attrs`) will be copied from
-        the original object to the new one.  If False (default), the new
-        object will be returned without attributes.
-    fill_value : scalar, optional
-        Value to fill DataArray(s) with before returning.
+    other : DataArray, Dataset, or Variable
+        The reference object in input
+    fill_value : scalar
+        Value to fill the new object with before returning it.
+    dtype : dtype, optional
+        dtype of the new array. If omitted, it defaults to other.dtype.
 
     Returns
     -------
-    out : same as xray_obj
-        New object with the same shape and type as a given object.
+    out : same as object
+        New object with the same shape and type as other, with the data
+        filled with fill_value. Coords will be copied from other.
+        If other is based on dask, the new one will be as well, and will be
+        split in the same chunks.
     """
     from .dataarray import DataArray
     from .dataset import Dataset
+    from .variable import Variable
 
-    if isinstance(xray_obj, Dataset):
-        attrs = xray_obj.attrs if keep_attrs else {}
+    if isinstance(other, Dataset):
+        data_vars = OrderedDict(
+            (k, _full_like_variable(v, fill_value, dtype))
+            for k, v in other.data_vars.items())
+        return Dataset(data_vars, coords=other.coords, attrs=other.attrs)
+    elif isinstance(other, DataArray):
+        return DataArray(
+            _full_like_variable(other.variable, fill_value, dtype),
+            dims=other.dims, coords=other.coords, attrs=other.attrs,
+            name=other.name)
+    elif isinstance(other, Variable):
+        return _full_like_variable(other, fill_value, dtype)
+    else:
+        raise TypeError("Expected DataArray, Dataset, or Variable")
 
-        return Dataset(dict((k, _full_like_dataarray(v, keep_attrs=keep_attrs,
-                                                     fill_value=fill_value))
-                            for k, v in iteritems(xray_obj.data_vars)),
-                       name=xray_obj.name, attrs=attrs)
-    elif isinstance(xray_obj, DataArray):
-        return _full_like_dataarray(xray_obj, keep_attrs=keep_attrs,
-                                    fill_value=fill_value)
+
+def _full_like_variable(other, fill_value, dtype=None):
+    """Inner function of full_like, where other must be a variable
+    """
+    from .variable import Variable
+
+    if isinstance(other.data, dask_array_type):
+        import dask.array
+        if dtype is None:
+            dtype = other.dtype
+        data = dask.array.full(other.shape, fill_value, dtype=dtype,
+                               chunks=other.data.chunks)
+    else:
+        data = np.full_like(other, fill_value, dtype=dtype)
+
+    return Variable(dims=other.dims, data=data, attrs=other.attrs)
+
+
+def zeros_like(other, dtype=None):
+    """Shorthand for full_like(other, 0, dtype)
+    """
+    return full_like(other, 0, dtype)
+
+
+def ones_like(other, dtype=None):
+    """Shorthand for full_like(other, 1, dtype)
+    """
+    return full_like(other, 1, dtype)
